export class NewTootModalOpen extends Error {}

export class NewTootBlockSubmit extends Error {}

export class NewTootTootLength extends Error {}

export class NewTootAttachLength extends Error {}

export class NewTootMediaDescription extends Error {}

export class NewTootPollInvalid extends Error {}

export class NewTootUnknownType extends Error {}

export class AuthenticationError extends Error {}
